@@ -29,6 +29,7 @@
 
 import os, time, zipfile, errno, datetime
 from StringIO import StringIO
+import tarfile
 
 from werkzeug import http_date
 
@@ -44,7 +45,6 @@ from MoinMoin.Page import Page
 from MoinMoin.util import filesys, timefuncs
 from MoinMoin.security.textcha import TextCha
 from MoinMoin.events import FileAttachedEvent, FileRemovedEvent, send_event
-from MoinMoin.support import tarfile
 
 action_name = __name__.split('.')[-1]
 
@@ -313,6 +313,40 @@ def move_attachment(request, pagename, dest_pagename, target, dest_target,
     return dest_target, filesize
 
 
+def copy_attachment(request, pagename, dest_pagename, target, dest_target,
+                    overwrite=False):
+    """ copy attachment <target> of page <pagename>
+        to attachment <dest_target> of page <dest_pagename>
+
+        note: this is lowlevel code, acl permissions need to be checked before
+              and also the target page should somehow exist (can be "deleted",
+              but the pagedir should be there)
+    """
+    # replace illegal chars
+    target = wikiutil.taintfilename(target)
+    dest_target = wikiutil.taintfilename(dest_target)
+
+    attachment_path = os.path.join(getAttachDir(request, pagename),
+                                   target).encode(config.charset)
+    dest_attachment_path = os.path.join(getAttachDir(request, dest_pagename, create=1),
+                                        dest_target).encode(config.charset)
+    if not overwrite and os.path.exists(dest_attachment_path):
+        raise DestPathExists
+    if dest_attachment_path == attachment_path:
+        raise SamePath
+    filesize = os.path.getsize(attachment_path)
+    try:
+        filesys.copy(attachment_path, dest_attachment_path)
+    except Exception:
+        raise
+    else:
+        _addLogEntry(request, 'ATTNEW', dest_pagename, dest_target)
+        event = FileAttachedEvent(request, dest_pagename, dest_target, filesize)
+        send_event(event)
+
+    return dest_target, filesize
+
+
 #############################################################################
 ### Internal helpers
 #############################################################################
@@ -460,7 +494,7 @@ function checkAll(bx, targets_name) {
                         links.append(fmt.url(1, getAttachUrl(pagename, file, request, do='unzip')) +
                                      fmt.text(label_unzip) +
                                      fmt.url(0))
-            except (RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile):
+            except (IOError, RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile):
                 # We don't want to crash with a traceback here (an exception
                 # here could be caused by an uploaded defective zip file - and
                 # if we crash here, the user does not get a UI to remove the
@@ -470,7 +504,8 @@ function checkAll(bx, targets_name) {
                 # archive).
                 # BadZipfile/LargeZipFile are raised when there are some
                 # specific problems with the archive file.
-                logging.exception("An exception within zip file attachment handling occurred:")
+                logging.exception("An exception within zip file attachment [%r:%r] handling occurred:" % (
+                     pagename, file))
 
             html.append(fmt.listitem(1))
             html.append("[%s]" % "&nbsp;| ".join(links))
@@ -483,6 +518,7 @@ function checkAll(bx, targets_name) {
 &nbsp;%(all_files)s&nbsp;|&nbsp;%(sel_files)s
 <input type="radio" name="multifile" value="rm">%(delete)s</input>
 <input type="radio" name="multifile" value="mv">%(move)s</input>
+<input type="radio" name="multifile" value="cp">%(copy)s</input>
 <input type="text" name="multi_dest_pagename" value="%(pagename)s">
 <input type="submit" value="%(submit)s">
 """ % dict(
@@ -490,7 +526,8 @@ function checkAll(bx, targets_name) {
             sel_files=_("Selected Files:"),
             delete=_("delete"),
             move=_("move to page"),
-            pagename=pagename,
+            copy=_("copy to page"),
+            pagename=wikiutil.escape(pagename),
             submit=_("Do it."),
 ))
         html.append("</form>")
@@ -509,6 +546,7 @@ def _do_multifile(pagename, request):
     _ = request.getText
     action = request.form.get('multifile')
     fnames = request.form.getlist('fn')
+    fails = []
     if action == 'rm':
         if not request.user.may.delete(pagename):
             return _('You are not allowed to delete attachments on this page.')
@@ -524,12 +562,34 @@ def _do_multifile(pagename, request):
         if not request.user.may.write(dest_pagename):
             return _('You are not allowed to attach a file to this page.')
         for fn in fnames:
-            move_attachment(request, pagename, dest_pagename, fn, fn)
+            try:
+                move_attachment(request, pagename, dest_pagename, fn, fn)
+            except (DestPathExists, SamePath):
+                fails.append(fn)
         msg = _("Attachment '%(pagename)s/%(filename)s' moved to '%(new_pagename)s/%(new_filename)s'.") % dict(
                 pagename=pagename,
                 filename=u'{%s}' % ','.join(fnames),
                 new_pagename=dest_pagename,
                 new_filename=u'*')
+        if fails:
+            msg += " " + _("Failed: %s", ", ".join(fails))
+        return upload_form(pagename, request, msg=msg)
+    if action == 'cp':
+        dest_pagename = request.form.get('multi_dest_pagename')
+        if not request.user.may.write(dest_pagename):
+            return _('You are not allowed to attach a file to this page.')
+        for fn in fnames:
+            try:
+                copy_attachment(request, pagename, dest_pagename, fn, fn)
+            except (DestPathExists, SamePath):
+                fails.append(fn)
+        msg = _("Attachment '%(pagename)s/%(filename)s' copied to '%(new_pagename)s/%(new_filename)s'.") % dict(
+                pagename=pagename,
+                filename=u'{%s}' % ','.join(fnames),
+                new_pagename=dest_pagename,
+                new_filename=u'*')
+        if fails:
+            msg += " " + _("Failed: %s", ", ".join(fails))
         return upload_form(pagename, request, msg=msg)
     return u'unsupported multifile operation'
 
@@ -1119,7 +1179,7 @@ def _do_unzip(pagename, request, overwrite=False):
                         'filelist': ', '.join(not_overwritten), }
             else:
                 msg = _("Attachment '%(filename)s' unzipped.") % {'filename': filename}
-    except (RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile), err:
+    except (IOError, RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile), err:
         # We don't want to crash with a traceback here (an exception
         # here could be caused by an uploaded defective zip file - and
         # if we crash here, the user does not get a UI to remove the
@@ -1129,7 +1189,8 @@ def _do_unzip(pagename, request, overwrite=False):
         # archive).
         # BadZipfile/LargeZipFile are raised when there are some
         # specific problems with the archive file.
-        logging.exception("An exception within zip file attachment handling occurred:")
+        logging.exception("An exception within zip file attachment [%r:%r] handling occurred:" % (
+            pagename, filename))
         msg = _("A severe error occurred:") + ' ' + str(err)
 
     upload_form(pagename, request, msg=msg)
@@ -1200,7 +1261,7 @@ def send_viewfile(pagename, request):
                 request.write(wikiutil.escape("%-46s %s %12d\n" % (zinfo.filename, date, zinfo.file_size)))
             request.write("</pre>")
             return
-    except (RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile):
+    except (IOError, RuntimeError, zipfile.BadZipfile, zipfile.LargeZipFile):
         # We don't want to crash with a traceback here (an exception
         # here could be caused by an uploaded defective zip file - and
         # if we crash here, the user does not get a UI to remove the
@@ -1210,7 +1271,8 @@ def send_viewfile(pagename, request):
         # archive).
         # BadZipfile/LargeZipFile are raised when there are some
         # specific problems with the archive file.
-        logging.exception("An exception within zip file attachment handling occurred:")
+        logging.exception("An exception within zip file attachment [%r:%r] handling occurred:" % (
+            pagename, filename))
         return
 
     from MoinMoin import macro
